/****************************************************************************/
/*!
\par Project:  Video for DC
\file          video_out_dc.c
\brief		   Dream Cast mpeg2 video renderer ... Alpha 0.1
\author        La Cible (la.cible@free.fr)
\date          Created on 17/10/2002 
\date          Updated on 17/10/2002 

*/
/****************************************************************************/
#include <kos.h>
#include "config.h"

#include <stdlib.h>
#include <stdio.h>
#include <inttypes.h>

#include "video_out.h"
#include "video_out_internal.h"

typedef struct dc_frame_s {
    vo_frame_t vo;
} dc_frame_t;

typedef struct dc_instance_s {
    vo_instance_t vo;
    int prediction_index;
    vo_frame_t * frame_ptr[3];
    dc_frame_t frame[3];
} dc_instance_t;

static int g_width=0;
static int g_height=0;

#define MATRIX_COEF 5

unsigned char* Clip;// Clip table 

int Inverse_Table_6_9[8][4]
=
{
	{117504, 138453, 13954, 34903}, /* no sequence_display_extension */
	{117504, 138453, 13954, 34903}, /* ITU-R Rec. 709 (1990) */
	{104597, 132201, 25675, 53279}, /* unspecified */
	{104597, 132201, 25675, 53279}, /* reserved */
	{104448, 132798, 24759, 53109}, /* FCC */
	{104597, 132201, 25675, 53279}, /* ITU-R Rec. 624-4 System B, G */
	{104597, 132201, 25675, 53279}, /* SMPTE 170M */
	{117579, 136230, 16907, 35559}  /* SMPTE 240M (1987) */
};

//CHROMA Conversion
static unsigned char *u422, *v422, *u444, *v444;
/* horizontal 1:2 interpolation filter */
void conv422to444(unsigned char *src,unsigned char *dst)
{
  int i, i2, w, j, im2, im1, ip1, ip2, ip3;

  w = g_width>>1;

  for (j=0; j<g_height; j++)
    {
      for (i=0; i<w; i++)
      {
        i2 = i<<1;
        im2 = (i<2) ? 0 : i-2;
        im1 = (i<1) ? 0 : i-1;
        ip1 = (i<w-1) ? i+1 : w-1;
        ip2 = (i<w-2) ? i+2 : w-1;
        ip3 = (i<w-3) ? i+3 : w-1;

        /* FIR filter coefficients (*256): 21 0 -52 0 159 256 159 0 -52 0 21 */
        /* even samples (0 0 256 0 0) */
        dst[i2] = src[i];

        /* odd samples (21 -52 159 159 -52 21) */
        dst[i2+1] = Clip[(int)(21*(src[im2]+src[ip3])
                        -52*(src[im1]+src[ip2]) 
                       +159*(src[i]+src[ip1])+128)>>8];
      }
      src+= w;
      dst+= g_width;
    }
 
}

/* vertical 1:2 interpolation filter */
void conv420to422(unsigned char *src,unsigned char *dst,dc_frame_t * frame)
{
  int w, h, i, j, j2;
  int jm6, jm5, jm4, jm3, jm2, jm1, jp1, jp2, jp3, jp4, jp5, jp6, jp7;

  w = g_width>>1;
  h = g_height>>1;

  if (frame->vo.progressive_frame)
  {
    /* intra frame */
    for (i=0; i<w; i++)
    {
      for (j=0; j<h; j++)
      {
        j2 = j<<1;
        jm3 = (j<3) ? 0 : j-3;
        jm2 = (j<2) ? 0 : j-2;
        jm1 = (j<1) ? 0 : j-1;
        jp1 = (j<h-1) ? j+1 : h-1;
        jp2 = (j<h-2) ? j+2 : h-1;
        jp3 = (j<h-3) ? j+3 : h-1;

        /* FIR filter coefficients (*256): 5 -21 70 228 -37 11 */
        /* New FIR filter coefficients (*256): 3 -16 67 227 -32 7 */
        dst[w*j2] =     Clip[(int)(  3*src[w*jm3]
                             -16*src[w*jm2]
                             +67*src[w*jm1]
                            +227*src[w*j]
                             -32*src[w*jp1]
                             +7*src[w*jp2]+128)>>8];

        dst[w*(j2+1)] = Clip[(int)(  3*src[w*jp3]
                             -16*src[w*jp2]
                             +67*src[w*jp1]
                            +227*src[w*j]
                             -32*src[w*jm1]
                             +7*src[w*jm2]+128)>>8];
      }
      src++;
      dst++;
    }
  }
  else
  {
    /* intra field */
    for (i=0; i<w; i++)
    {
      for (j=0; j<h; j+=2)
      {
        j2 = j<<1;

        /* top field */
        jm6 = (j<6) ? 0 : j-6;
        jm4 = (j<4) ? 0 : j-4;
        jm2 = (j<2) ? 0 : j-2;
        jp2 = (j<h-2) ? j+2 : h-2;
        jp4 = (j<h-4) ? j+4 : h-2;
        jp6 = (j<h-6) ? j+6 : h-2;

        /* Polyphase FIR filter coefficients (*256): 2 -10 35 242 -18 5 */
        /* New polyphase FIR filter coefficients (*256): 1 -7 30 248 -21 5 */
        dst[w*j2] = Clip[(int)(  1*src[w*jm6]
                         -7*src[w*jm4]
                         +30*src[w*jm2]
                        +248*src[w*j]
                         -21*src[w*jp2]
                          +5*src[w*jp4]+128)>>8];

        /* Polyphase FIR filter coefficients (*256): 11 -38 192 113 -30 8 */
        /* New polyphase FIR filter coefficients (*256):7 -35 194 110 -24 4 */
        dst[w*(j2+2)] = Clip[(int)( 7*src[w*jm4]
                             -35*src[w*jm2]
                            +194*src[w*j]
                            +110*src[w*jp2]
                             -24*src[w*jp4]
                              +4*src[w*jp6]+128)>>8];

        /* bottom field */
        jm5 = (j<5) ? 1 : j-5;
        jm3 = (j<3) ? 1 : j-3;
        jm1 = (j<1) ? 1 : j-1;
        jp1 = (j<h-1) ? j+1 : h-1;
        jp3 = (j<h-3) ? j+3 : h-1;
        jp5 = (j<h-5) ? j+5 : h-1;
        jp7 = (j<h-7) ? j+7 : h-1;

        /* Polyphase FIR filter coefficients (*256): 11 -38 192 113 -30 8 */
        /* New polyphase FIR filter coefficients (*256):7 -35 194 110 -24 4 */
        dst[w*(j2+1)] = Clip[(int)( 7*src[w*jp5]
                             -35*src[w*jp3]
                            +194*src[w*jp1]
                            +110*src[w*jm1]
                             -24*src[w*jm3]
                              +4*src[w*jm5]+128)>>8];

        dst[w*(j2+3)] = Clip[(int)(  1*src[w*jp7]
                             -7*src[w*jp5]
                             +30*src[w*jp3]
                            +248*src[w*jp1]
                             -21*src[w*jm1]
                              +5*src[w*jm3]+128)>>8];
      }
      src++;
      dst++;
    }
  }
}


static void dc_draw_frame (vo_frame_t * frame)
{

	int i,j;
	int y, u, v, r, g, b;
	int crv, cbu, cgu, cgv;
	unsigned char *py, *pu, *pv;
	static int k = 0;

	if((k++%3)!=0)
		return;
	
	dc_frame_t * pframe = (dc_frame_t *) frame;

	conv420to422(pframe->vo.base[1],u422,pframe);
	conv420to422(pframe->vo.base[2],v422,pframe);
	conv422to444(u422,u444);
	conv422to444(v422,v444);

  crv = Inverse_Table_6_9[MATRIX_COEF][0];
  cbu = Inverse_Table_6_9[MATRIX_COEF][1];
  cgu = Inverse_Table_6_9[MATRIX_COEF][2];
  cgv = Inverse_Table_6_9[MATRIX_COEF][3];

  for (i=0; i<g_height; i++)
  {
    py = pframe->vo.base[0]  + g_width*i;
    pu = u444  + g_width*i;
    pv = v444  + g_width*i;

    for (j=0; j<g_width; j++)
    {
      u = *pu++ - 128;
      v = *pv++ - 128;
      y = 76309 * (*py++ - 16); /* (255/219)*65536 */
      r = Clip[(y + crv*v + 32768)>>16];
      g = Clip[(y - cgu*u - cgv*v + 32768)>>16];
      b = Clip[(y + cbu*u + 32786)>>16];

	  r = (r * (1<<4))/255;
      g = (g * (1<<4))/255;
	  b = (b * (1<<4))/255;
	/********************************
	 * Flush into video buffer
	 ********************************/
	 vram_s[i*g_width+j] =(r << 10)
				| (g << 5)
				| (b << 0);
       

    }
  }


}


static vo_frame_t * rgb_get_frame (vo_instance_t * _instance, int flags)
{
    dc_instance_t * instance;
    dc_frame_t * frame;

    instance = (dc_instance_t *) _instance;
    frame =
	(dc_frame_t *) libvo_common_get_frame ((vo_instance_t *) instance,
						 flags);


    return (vo_frame_t *) frame;
}


static int dcrgb_setup (vo_instance_t * _instance, int width, int height)
{
    dc_instance_t * instance;

    instance = (dc_instance_t *) _instance;
	g_width=width;
	g_height=height;
	/* Set the video mode */
	if(g_width==320 && g_height==240) 
		vid_set_mode(DM_320x240, PM_RGB555);
	else if(g_width==640 && g_height==480) 
		vid_set_mode(DM_640x480, PM_RGB555);
	else
	{
		printf("[Video For DreamCast]\nError resolution not supported\n");
		return 0;
	}
//	printf("VIDEO resolution %dx%d\n",width,height);

	u422 = (unsigned char *)malloc((width>>1)*height);
		
	v422 = (unsigned char *)malloc((width>>1)*height);
	
	u444 = (unsigned char *)malloc(width*height);
	
	v444 = (unsigned char *)malloc(width*height);

		
 

    return libvo_common_alloc_frames ((vo_instance_t *) instance,
				      width, height, sizeof (dc_frame_t),
				      NULL,NULL,
				      dc_draw_frame);
}

void vo_dc_close(vo_instance_t * _instance)
{
	free(u422);
	free(v422);
	free(u444);
	free(v444);
	free(Clip);

	libvo_common_free_frames(_instance);
}

vo_instance_t * vo_dc_open (void)
{
    dc_instance_t * instance;
	int i;

    instance = malloc (sizeof (dc_instance_t));
    if (instance == NULL)
	return NULL;

	if (!(Clip=(unsigned char *)malloc(1024)))
	{
		printf("Clip[] malloc failed\n");
		return NULL;
	}
	
	Clip += 384;
	
	for (i=-384; i<640; i++)
		Clip[i] = (i<0) ? 0 : ((i>255) ? 255 : i);
	
	

    instance->vo.setup = dcrgb_setup;
    instance->vo.close = vo_dc_close;//libvo_common_free_frames;
    instance->vo.get_frame = rgb_get_frame;
   
    return (vo_instance_t *) instance;
}

